#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"ureg.h"
#include	"../port/error.h"

enum
{
	SE_WRITE	= 4<<5,
	SE_PROT		= 2<<2,
};

void
faultsparc(Ureg *ur)
{
	ulong addr, badvaddr;
	char buf[ERRLEN];
	int user, read;
	ulong tbr, ser;

	tbr = (ur->tbr&0xFFF)>>4;
	addr = ur->pc;			/* assume instr. exception */
	read = 1;
	if(tbr == 9){			/* data access exception */
		addr = getrmmu(SFAR);
		ser = getrmmu(SFSR);
		if(ser&(SE_WRITE))	/* is SE_PROT needed? */
			read = 0;
	}

	spllo();
	if(u == 0){
		dumpregs(ur);
		panic("fault u==0 pc=%lux addr=%lux", ur->pc, addr);
	}

	badvaddr = addr;
	addr &= ~(BY2PG-1);
	user = !(ur->psr&PSRPSUPER);
	if(fault(addr, read) < 0){
		if(user){
			sprint(buf, "sys: trap: fault %s addr=0x%lux",
				read? "read" : "write", badvaddr);
			postnote(u->p, 1, buf, NDebug);
			return;
		}
		dumpregs(ur);
		panic("fault: 0x%lux", badvaddr);
	}
	splhi();
}

void
faultasync(Ureg *ur)
{
	int user;

	print("interrupt 15 AFSR %lux AFAR %lux MFSR %lux MFAR %lux\n",
		getphys(AFSR), getphys(AFAR), getphys(MFSR), getphys(MFAR));
	dumpregs(ur);
	/*
	 * Clear interrupt
	 */
	putphys(PROCINTCLR, 1<<15);
	user = !(ur->psr&PSRPSUPER);
	if(user)
		pexit("Suicide", 0);
	panic("interrupt 15");
}

/*
 * called in sysfile.c
 */
void
evenaddr(ulong addr)
{
	if(addr & 3){
		postnote(u->p, 1, "sys: odd address", NDebug);
		error(Ebadarg);
	}
}
