#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"init.h"

#include	<libg.h>
#include	<gnot.h>

Sysint	*sysintr;
struct {
	uchar	format;
	uchar	type;
	uchar	ea[6];
	uchar	pad[32-8];
} idprom;
ulong	rom;		/* open boot rom vector */
int	cpuserver;
ulong	bank[8];
uchar	mempres[64];
char	fbstr[32];
ulong	fbslot;
int	usecg6;
Label	catch;
uchar	*sp;

typedef struct Sysparam Sysparam;
struct Sysparam
{
	int	id;		/* Model type from id prom */
	char	*name;		/* System name */
	char	ss2;		/* Is Sparcstation 2? */
	int	vacsize;	/* Cache size */
	int	vacline;	/* Cache line size */
	int	ncontext;	/* Number of MMU contexts */
	char	cachebug;	/* Machine needs cache bug work around */
	int	nbank;		/* Number of banks of memory */
	int	banksize;	/* Maximum Mbytes per bank */
}
sysparam[] =
{
	{ 0xFF, "unknown Sun4M",0, 0, 0,  64, 0, 4, 32 },
	{ 0x80, "TMS390S10",0, 0, 0,  64, 0, 4, 32 },
	{ 0 }
};
Sysparam *sparam;

void
main(void)
{
	u = 0;
	memset(edata, 0, (char*)end-(char*)edata);

	machinit();
	active.exiting = 0;
	active.machs = 1;
	trapinit();
	confinit();
	xinit();
	mmuinit();
	kmapinit();
	if(conf.monitor)
		screeninit(fbstr, fbslot);
	printinit();
	ioinit();
	if(conf.monitor == 0)
		sccspecial(2, &printq, &kbdq, 9600);
	pageinit();
	cacheinit();
	intrinit();
	procinit0();
	initseg();
putphys(MID, 0x1F<<16);	/* enable arbitration */
	chandevreset();
	streaminit();
	swapinit();
	userinit();
	clockinit();
	putphys(PROCINTCLR, (~0<<17)|(1<<15));	/* clear pending processor interrupts */
	schedinit();
}

void
intrinit(void)
{
	KMap *k;

	getphys(AFSR);	/* clear fault status */
	k = kmappa(SYSINTR, PTEIO|PTENOCACHE);
	sysintr = (Sysint*)VA(k);
	sysintr->maskset = ~0;	/* mask all interrupts */
	sysintr->maskclr = MaskAllIntr|MEIntr|MSIIntr|EMCIntr|ScsiIntr|EtherIntr|SCCIntr|KbdIntr; /* allow these */
	putphys(PROCINTCLR, (~0<<17)|(1<<15));	/* clear pending processor interrupts */
}

void
systemreset(void)
{
	delay(200);
	putphys(SYSCTL, getphys(SYSCTL)|1);	/* power on reset */
}


void
machinit(void)
{
	int n;

	n = m->machno;
	memset(m, 0, sizeof(Mach));
	m->machno = n;
	m->mmask = 1<<m->machno;
	m->fpstate = FPinit;
}

void
ioinit(void)
{
	KMap *k;

	/* tell scc driver its addresses */
	k = kmappa(KMDUART, PTEIO|PTENOCACHE);
	sccsetup((void*)VA(k), KMFREQ, 1);
	k = kmappa(EIADUART, PTEIO|PTENOCACHE);
	sccsetup((void*)VA(k), EIAFREQ, 1);

	/* scc port 0 is the keyboard */
	sccspecial(0, 0, &kbdq, 2400);
	kbdq.putc = kbdstate;

	/* scc port 1 is the mouse */
	sccspecial(1, 0, &mouseq, 2400);
}

void
init0(void)
{
	u->nerrlab = 0;
	m->proc = u->p;
	u->p->state = Running;
	u->p->mach = m;
	spllo();

	print("Sun Sparc %s\n", sparam->name);
	print("bank 0: %dM  1: %dM\n", bank[0], bank[1]);
	print("frame buffer id %lux slot %ld %s\n", conf.monitor, fbslot, fbstr);

	u->slash = (*devtab[0].attach)(0);
	u->dot = clone(u->slash, 0);

	kproc("alarm", alarmkproc, 0);
	chandevinit();

	if(!waserror()){
		ksetterm("sun %s");
		ksetenv("cputype", "sparc");
		poperror();
	}
	touser((long)sp);
}

FPsave	*initfpp;
uchar	initfpa[sizeof(FPsave)+7];

void
userinit(void)
{
	Proc *p;
	Segment *s;
	User *up;
	KMap *k;
	ulong l;
	Page *pg;

	p = newproc();
	p->pgrp = newpgrp();
	p->egrp = smalloc(sizeof(Egrp));
	p->egrp->ref = 1;
	p->fgrp = smalloc(sizeof(Fgrp));
	p->fgrp->ref = 1;
	p->procmode = 0640;

	strcpy(p->text, "*init*");
	strcpy(p->user, eve);
	/* must align initfpp to an ODD word boundary */
	l = (ulong)initfpa;
	l += 3;
	l &= ~7;
	l += 4;
	initfpp = (FPsave*)l;
	savefpregs(initfpp);
	p->fpstate = FPinit;

	/*
	 * Kernel Stack
	 */
	p->sched.pc = (((ulong)init0) - 8);	/* 8 because of RETURN in gotolabel */
	p->sched.sp = USERADDR+BY2PG-(1+MAXSYSARG)*BY2WD;
	p->sched.sp &= ~7;		/* SP must be 8-byte aligned */
	p->upage = newpage(1, 0, USERADDR|(p->pid&0xFFFF));

	/*
	 * User
	 */
	k = kmap(p->upage);
	up = (User*)VA(k);
	up->p = p;
	kunmap(k);

	/*
	 * User Stack
	 */
	s = newseg(SG_STACK, USTKTOP-BY2PG, 1);
	p->seg[SSEG] = s;
	pg = newpage(1, 0, USTKTOP-BY2PG);
	segpage(s, pg);
	k = kmap(pg);
	bootargs(VA(k));
	kunmap(k);

	/*
	 * Text
	 */
	s = newseg(SG_TEXT, UTZERO, 1);
	p->seg[TSEG] = s;
	segpage(s, newpage(1, 0, UTZERO));
	k = kmap(s->map[0]->pages[0]);
	memmove((ulong*)VA(k), initcode, sizeof initcode);
	kunmap(k);

	ready(p);
}

uchar *
pusharg(char *p)
{
	int n;

	n = strlen(p)+1;
	sp -= n;
	memmove(sp, p, n);
	return sp;
}

void
bootargs(ulong base)
{
 	int i, ac;
	uchar *av[32];
	uchar **lsp;

	sp = (uchar*)base + BY2PG - MAXSYSARG*BY2WD;

	ac = 0;
	av[ac++] = pusharg("/sparc/9ss");
	av[ac++] = pusharg("-p");

	/* 4 byte word align stack */
	sp = (uchar*)((ulong)sp & ~3);

	/* build argc, argv on stack */
	sp -= (ac+1)*sizeof(sp);
	lsp = (uchar**)sp;
	for(i = 0; i < ac; i++)
		*lsp++ = av[i] + ((USTKTOP - BY2PG) - base);
	*lsp = 0;
	sp += (USTKTOP - BY2PG) - base - sizeof(sp);
}

void
exit(int ispanic)
{
	int i;

	u = 0;
	spllo();
	print("cpu %d exiting\n", m->machno);
	while(consactive())
		for(i=0; i<1000; i++)
			;
	splhi();
	if(ispanic)
		for(;;);
	systemreset();
}

int
probemem(ulong addr)
{
	ulong pcr, save0;
	int works;

	save0 = getphys(0);
	pcr = getpcr()|NOFAULT;
	works = 0;
	setpcr(pcr & ~MEMPCHECK);
	putphys(addr, ~addr);
	if(addr)
		putphys(0, 0x89ABCDEF);
	if(getphys(addr) == ~addr){
		setpcr(pcr);
		putphys(addr, addr);
		if(addr)
			putphys(0, 0x89ABCDEF);
		if(getphys(addr) == addr)
			works = 1;
	}
	setpcr(pcr & ~NOFAULT);
	putphys(0, save0);
	getphys(AFSR);	/* clear fault status */
	getrmmu(SFSR);	/* clear fault status */
	return works;
}

/*
 * this assumes that if a bank is not empty,
 * its first slot is filled.
 *
 * /sys/src/9/port/alloc.c and /sys/src/9/port/page.c
 * need to be changed to support more than two banks.
 */
void
scanbank(ulong base, uchar *mempres, int n)
{
	int i;
	ulong addr, npg;

	npg = 0;
	for(i=0; i<n; i++){
		mempres[i] = 0;
		addr = base + i*MB;
		if(!probemem(addr))
			break;
		if(addr != base) {
			/* check for mirrors */
			putphys(addr, addr);
			if(getphys(base) == addr)
				break;
		}
		mempres[i] = 1;
		npg += MB/BY2PG;
	}
	if(npg){
		if(conf.npage0 == 0){
			conf.base0 = base;
			conf.npage0 = npg;
		}else if(conf.npage1 < npg){
			conf.base1 = base;
			conf.npage1 = npg;
		}
	}
}

void
physcopyin(void *d, ulong s, int n)
{
	int i, j;
	ulong w;

	for(i=0; i<n; i+=sizeof(ulong)) {
		w = getphys(s+i);
		j = n-i;
		if(j > sizeof(ulong))
			j = sizeof(ulong);
		memmove((uchar*)d+i, &w, j);
	}
}

Conf	conf;

void
confinit(void)
{
	int mul;
	ulong i, j, f;
	ulong ktop, pa, mbytes;
	char buf[8];

	conf.nmach = 1;
	if(conf.nmach > MAXMACH)
		panic("confinit");

	/* fetch ID prom */
	physcopyin(&idprom, NVRAM+IDOFF, sizeof(idprom));
	if(idprom.format!=1 || (idprom.type&0xF0)!=0x80)
		*(ulong*)~0 = 0;	/* not a new generation sparc; die! */
	/*
	 * since Sun seems to have changed things,
	 * making it hard to probe for the video type without resetting the video adapter,
	 * temporarily distinguish Classic (cgthree) from LX (cgsix)
	 * by checking for the DBRI in SBUS virtual slot 2,
	 * assuming that only the LX has the thing.
	 */
	conf.monitor = 1;
	usecg6 = 0;
	if(setlabel(&catch) == 0){ /* LX has DBRI in slot 2 */
		if(getphys(SBUS(2))!=0)
			usecg6 = 1;
	}
	catch.pc = 0;
	getphys(AFSR);	/* clear error status */
	strcpy(fbstr, usecg6? "cgsix": "cgthree");
	fbslot = 3;
/*
 * scanning for a frame buffer this way doesn't work on the Classic/LX:
 * touching the frame buffer id area safely used on the ss2 now resets the device
 */
	/*
	 * Look for a frame buffer.  Do it the way the
	 * ROM does it: scan the slots in a specified order and use
	 * the first one it finds.
	 *
	 * If we find a frame buffer, we always use it as a console
	 * rather than the attached terminal, if any.  This means
	 * if you have a frame buffer you'd better have a builtin
	 * keyboard, too.
	 */
	/* map frame buffer id in SBUS slots 0..3 */
	for(f=0; f<=3 && fbstr[0]==0; f++){
		pa = FRAMEBUFID(f);
		conf.monitor = 0;
		/* if frame buffer not present, we will trap, so prepare to catch it */
		if(setlabel(&catch) == 0){
			conf.monitor = getphys(pa);
			switch(conf.monitor & 0xF0FFFFFF){
			case 0xF0010101:
				strcpy(fbstr, "cgthree");
				break;
			case 0xF0010104:
				strcpy(fbstr, "bwtwo");
				break;
			}
			if(fbstr[0] == 0){
				j = getphys(pa+4);
				if(j > BY2PG-sizeof(buf))
					j = BY2PG - sizeof(buf);	/* for safety */
				for(i=0; i<j && fbstr[0]==0; i++) {
					physcopyin(buf, pa+i, sizeof(buf));
					if(strncmp(buf, "bwtwo", 5) == 0)
						strcpy(fbstr, "bwtwo");
					if(strncmp(buf, "cgthree", 7) == 0)
						strcpy(fbstr, "cgthree");
					if(strncmp(buf, "cgsix", 5) == 0)
						strcpy(fbstr, "cgsix");
				}
			}
		}
		catch.pc = 0;
		if(fbstr[0])
			fbslot = f;
	}
	getphys(AFSR);	/* clear error status */

	for(sparam = sysparam; sparam->id; sparam++)
		if(sparam->id == idprom.type)
			break;

	/* First entry in the table is the default */
	if(sparam->id == 0)
		sparam = sysparam;

	conf.ss2 = sparam->ss2;
	conf.vacsize = sparam->vacsize;
	conf.vaclinesize = sparam->vacline;
	conf.ncontext = sparam->ncontext;
	conf.ss2cachebug = sparam->cachebug;

	for(i=0; i<sparam->nbank; i++)
		if(probemem(i*sparam->banksize*MB))
			scanbank(i*sparam->banksize*MB, mempres, sparam->banksize);

	bank[0] = conf.npage0*BY2PG/MB;
	bank[1] = conf.npage1*BY2PG/MB;

	if(bank[1] == 0){
		/*
		 * This split of memory into 2 banks fools the allocator into
		 * allocating low memory pages from bank 0 for the ethernet since
		 * it has only a 24bit address counter.
		 * NB. Suns must have at LEAST 8Mbytes.
		 */
		conf.npage1 = conf.npage0 - (8*MB)/BY2PG;
		conf.base1 = conf.base0 + 8*MB;
		conf.npage0 = (8*MB)/BY2PG;
		bank[1] = bank[0]-8;
		bank[0] = 8;
	}

	conf.npage = conf.npage0+conf.npage1;
	i = screenbits()-1;		/* Calculate % of memory for page pool */
	i = 70 - (i*10);
	conf.upages = (conf.npage*i)/100;
	if(cpuserver){
		i = conf.npage-conf.upages;
		if(i > (12*MB)/BY2PG)
			conf.upages +=  i - ((12*MB)/BY2PG);
	}


	ktop = PGROUND((ulong)end);
	ktop = PADDR(ktop);
	conf.npage0 -= ktop/BY2PG;
	conf.base0 += ktop;

	mbytes = (conf.npage*BY2PG)>>20;
	mul = 1 + (mbytes+11)/12;
	if(mul > 2)
		mul = 2;

	conf.nproc = 50*mul;
	if(cpuserver)
		conf.nswap = conf.npage*2;
	else
		conf.nswap = 16*MB/BY2PG;
	conf.nimage = 50;
	conf.copymode = 0;		/* copy on write */
	conf.arp = 32;
	if(cpuserver)
		conf.nproc = 500;
}

/*
 *  set up the lance
 */
void
lancesetup(Lance *lp)
{
	KMap *k;
	DMAdev *dma;
	ulong pa, va;
	int i;

	k = kmappa(ETHER, PTEIO|PTENOCACHE);
	lp->rdp = (void*)(VA(k)+0);
	lp->rap = (void*)(VA(k)+2);
	for(i=0; i<6; i++)
		lp->ea[i] = idprom.ea[i];
	

	lp->lognrrb = 7;
	lp->logntrb = 7;
	lp->nrrb = 1<<lp->lognrrb;
	lp->ntrb = 1<<lp->logntrb;
	lp->sep = 1;
	lp->busctl = BSWP | ACON | BCON;

	/*
	 * Allocate area for lance init block and descriptor rings
	 */
	pa = PADDR(xspanalloc(BY2PG, BY2PG, 0));

	/* map at LANCESEGM */
	va = kmapdma(pa, BY2PG);
	lp->lanceram = (ushort*)va;
	lp->lm = (Lancemem*)va;

	/*
	 * Allocate space in host memory for the io buffers.
	 */
	i = (lp->nrrb+lp->ntrb)*sizeof(Etherpkt);
	i = (i+(BY2PG-1))/BY2PG;
	pa = PADDR(xspanalloc(i*BY2PG, BY2PG, 0));
	va = kmapdma(pa, i*BY2PG);

	lp->lrp = (Etherpkt*)va;
	lp->rp = (Etherpkt*)va;
	lp->ltp = lp->lrp+lp->nrrb;
	lp->tp = lp->rp+lp->nrrb;

	k = kmappa(DMA, PTEIO|PTENOCACHE);
	dma = (DMAdev*)VA(k);
	dma->base = 0xff;
	/* for now, let's assume the ROM has left the results of its auto-sensing */
	if(dma->ecsr & E_TP_select)
		print("Twisted pair ethernet\n");
	else
		print("AUI ethernet\n");
	delay(1);
	dma->ecsr |= E_Int_en|E_Invalidate|E_Dsbl_wr_inval|E_Dsbl_rd_drn;
	delay(1);
}
